@@ -13,13 +13,15 @@
 #
 
 import os, sys, platform, shutil, _jcc
+from itertools import izip
 
 from cpp import PRIMITIVES, INDENT, HALF_INDENT
 from cpp import cppname, cppnames, typename
 from cpp import line, signature, find_method, split_pkg, sort
 from cpp import Modifier, Class, Method
 from _jcc import findClass
-from config import INCLUDES, CFLAGS, DEBUG_CFLAGS, LFLAGS, SHARED
+from config import INCLUDES, CFLAGS, DEBUG_CFLAGS, LFLAGS, IMPLIB_LFLAGS, \
+    SHARED, VERSION as JCC_VER
 
 try:
     from cpp import ParameterizedType, TypeVariable
@@ -32,7 +34,7 @@ if python_ver < '2.4':
 
 
 RESULTS = { 'boolean': 'Py_RETURN_BOOL(%s);',
-            'byte': 'return PyInt_FromLong(%s);',
+            'byte': 'return PyInt_FromLong((long) %s);',
             'char': 'return PyUnicode_FromUnicode((Py_UNICODE *) &%s, 1);',
             'double': 'return PyFloat_FromDouble((double) %s);',
             'float': 'return PyFloat_FromDouble((double) %s);',
@@ -51,10 +53,59 @@ CALLARGS = { 'boolean': ('O', '(%s ? Py_True : Py_False)', False),
              'short': ('i', '(int) %s', False),
              'java.lang.String': ('O', 'env->fromJString((jstring) %s, 0)', True) }
 
+BOXED = { 'java.lang.Boolean': (True, True),
+          'java.lang.Byte': (True, True),
+          'java.lang.Character': (True, True),
+          'java.lang.CharSequence': (True, False),
+          'java.lang.Double': (True, True),
+          'java.lang.Float': (True, True),
+          'java.lang.Integer': (True, True),
+          'java.lang.Long': (True, True),
+          'java.lang.Number': (True, False),
+          'java.lang.Short': (True, True),
+          'java.lang.String': (True, True) }
 
-def parseArgs(params, current, generics):
 
-    def signature(cls):
+def is_boxed(clsName):
+    return BOXED.get(clsName, (False, False))[0]
+
+def is_unboxed(clsName):
+    return BOXED.get(clsName, (False, False))[1]
+
+
+def getTypeParameters(cls):
+
+    while True:
+        parameters = cls.getTypeParameters()
+        if parameters:
+            return parameters
+        cls = cls.getDeclaringClass()
+        if cls is None:
+            return []
+
+
+def getActualTypeArguments(pt):
+
+    while True:
+        arguments = pt.getActualTypeArguments()
+        if arguments:
+            return arguments
+        pt = pt.getOwnerType()
+        if pt is None or not ParameterizedType.instance_(pt):
+            return []
+        pt = ParameterizedType.cast_(pt)
+
+
+def parseArgs(params, current, generics, genericParams=None):
+
+    def signature(cls, genericPT=None):
+        if generics and TypeVariable.instance_(genericPT):
+            if cls.getName() == 'java.lang.Object':
+                gd = TypeVariable.cast_(genericPT).getGenericDeclaration()
+                if gd == current:
+                    for clsParam in getTypeParameters(gd):
+                        if genericPT == clsParam:
+                            return 'O'
         array = ''
         while cls.isArray():
             array += '['
@@ -66,30 +117,53 @@ def parseArgs(params, current, generics):
             return array + 's'
         if clsName == 'java.lang.Object':
             return array + 'o'
-        if generics and cls.getTypeParameters():
+        if is_boxed(clsName):
+            return array + 'O'
+        if generics and getTypeParameters(cls):
             return array + 'K'
         else:
             return array + 'k'
 
-    def checkarg(cls):
+    def checkarg(cls, genericPT=None):
+        if generics and TypeVariable.instance_(genericPT):
+            if cls.getName() == 'java.lang.Object':
+                gd = TypeVariable.cast_(genericPT).getGenericDeclaration()
+                if gd == current:
+                    i = 0
+                    for clsParam in getTypeParameters(gd):
+                        if genericPT == clsParam:
+                            return ', self->parameters[%d]' %(i)
+                        i += 1
         while cls.isArray():
             cls = cls.getComponentType()
+        clsName = cls.getName()
         if (cls.isPrimitive() or
-            cls.getName() in ('java.lang.String', 'java.lang.Object')):
+            clsName in ('java.lang.String', 'java.lang.Object')):
             return ''
+        if is_boxed(clsName):
+            clsNames = clsName.split('.')
+            return ', &%s::PY_TYPE(%s)' %('::'.join(cppnames(clsNames[:-1])), cppname(clsNames[-1]))
         return ', %s::initializeClass' %(typename(cls, current, False))
 
     def callarg(cls, i):
         if generics:
             while cls.isArray():
                 cls = cls.getComponentType()
-            if cls.getTypeParameters():
+            if getTypeParameters(cls):
                 ns, sep, n = rpartition(typename(cls, current, False), '::')
                 return ', &a%d, &p%d, %s%st_%s::parameters_' %(i, i, ns, sep, n)
         return ', &a%d' %(i)
 
-    return (''.join([signature(param) for param in params]),
-            ''.join([checkarg(param) for param in params]),
+    if genericParams:
+        sig = ''.join([signature(param, genericParam)
+                       for param, genericParam in izip(params, genericParams)])
+        chk = ''.join([checkarg(param, genericParam)
+                       for param, genericParam in izip(params, genericParams)])
+    else:
+        sig = ''.join([signature(param) for param in params])
+        chk = ''.join([checkarg(param) for param in params])
+
+    return (sig, chk,
             ''.join([callarg(params[i], i) for i in xrange(len(params))]))
 
 
@@ -103,7 +177,7 @@ def declareVars(out, indent, params, current, generics, typeParams):
         if generics:
             while param.isArray():
                 param = param.getComponentType()
-            if param.getTypeParameters():
+            if getTypeParameters(param):
                 line(out, indent, 'PyTypeObject **p%d;', i)
                 typeParams.add(i)
     
@@ -212,16 +286,16 @@ def returnValue(cls, returnType, value, genericRT=None, typeParams=None):
         if ParameterizedType.instance_(genericRT):
             genericRT = ParameterizedType.cast_(genericRT)
             clsArgs = []
-            for clsArg in genericRT.getActualTypeArguments():
+            for clsArg in getActualTypeArguments(genericRT):
                 if Class.instance_(clsArg):
                     clsNames = Class.cast_(clsArg).getName().split('.')
-                    clsArg = '&%s::%s$$Type' %('::'.join(cppnames(clsNames[:-1])), cppname(clsNames[-1]))
+                    clsArg = '&%s::PY_TYPE(%s)' %('::'.join(cppnames(clsNames[:-1])), cppname(clsNames[-1]))
                     clsArgs.append(clsArg)
                 elif TypeVariable.instance_(clsArg):
                     gd = TypeVariable.cast_(clsArg).getGenericDeclaration()
                     if Class.instance_(gd):
                         i = 0
-                        for clsParam in gd.getTypeParameters():
+                        for clsParam in getTypeParameters(gd):
                             if clsArg == clsParam:
                                 clsArgs.append('self->parameters[%d]' %(i))
                                 break
@@ -234,16 +308,17 @@ def returnValue(cls, returnType, value, genericRT=None, typeParams=None):
                     break
             else:
                 return 'return %s%st_%s::wrap_Object(%s, %s);' %(ns, sep, n, value, ', '.join(clsArgs))
+
         elif TypeVariable.instance_(genericRT):
             gd = TypeVariable.cast_(genericRT).getGenericDeclaration()
             i = 0
             if Class.instance_(gd):
-                for clsParam in gd.getTypeParameters():
+                for clsParam in getTypeParameters(gd):
                     if genericRT == clsParam:
                         return 'return self->parameters[%d] != NULL ? wrapType(self->parameters[%d], %s.this$) : %s%st_%s::wrap_Object(%s);' %(i, i, value, ns, sep, n, value)
                     i += 1
             elif Method.instance_(gd):
-                for clsParam in gd.getTypeParameters():
+                for clsParam in getTypeParameters(gd):
                     if genericRT == clsParam and i in typeParams:
                         return 'return p%d != NULL && p%d[0] != NULL ? wrapType(p%d[0], %s.this$) : %s%st_%s::wrap_Object(%s);' %(i, i, i, value, ns, sep, n, value)
                     i += 1
@@ -264,9 +339,11 @@ def call(out, indent, cls, inCase, method, names, cardinality, isExtension,
     returnType = method.getReturnType()
     if generics:
         genericRT = method.getGenericReturnType()
+        genericParams = method.getGenericParameterTypes()
         typeParams = set()
     else:
         genericRT = None
+        genericParams = None
         typeParams = None
     count = len(params)
 
@@ -288,7 +365,7 @@ def call(out, indent, cls, inCase, method, names, cardinality, isExtension,
             line(out, indent, 'if (arg)')
         else:
             line(out, indent, 'if (!parseArg%s(arg%s, "%s"%s%s))',
-                 s, s, *parseArgs(params, cls, generics))
+                 s, s, *parseArgs(params, cls, generics, genericParams))
         line(out, indent, '{')
         indent += 1
 
@@ -304,7 +381,7 @@ def call(out, indent, cls, inCase, method, names, cardinality, isExtension,
     if isExtension and name == 'clone' and Modifier.isNative(modifiers):
         line(out)
         line(out, indent, '%s object(result.this$);', typename(cls, cls, False))
-        line(out, indent, 'if (PyObject_TypeCheck(arg, &FinalizerProxy$$Type) &&')
+        line(out, indent, 'if (PyObject_TypeCheck(arg, &PY_TYPE(FinalizerProxy)) &&')
         line(out, indent, '    PyObject_TypeCheck(((t_fp *) arg)->object, self->ob_type))')
         line(out, indent, '{')
         line(out, indent + 1, 'PyObject *_arg = ((t_fp *) arg)->object;')
@@ -465,7 +542,8 @@ def extension(env, out, indent, cls, names, name, count, method, generics):
 
 def python(env, out_h, out, cls, superCls, names, superNames,
            constructors, methods, protectedMethods, fields, instanceFields,
-           mapping, sequence, rename, declares, typeset, moduleName, generics):
+           mapping, sequence, rename, declares, typeset, moduleName, generics,
+           _dll_export):
 
     line(out_h)
     line(out_h, 0, '#include <Python.h>')
@@ -475,15 +553,16 @@ def python(env, out_h, out, cls, superCls, names, superNames,
     for name in names[:-1]:
         line(out_h, indent, 'namespace %s {', cppname(name))
         indent += 1
-    line(out_h, indent, 'extern PyTypeObject %s$$Type;', names[-1])
+    line(out_h, indent, '%sextern PyTypeObject PY_TYPE(%s);', 
+         _dll_export, names[-1])
 
     if generics:
-        clsParams = cls.getTypeParameters()
+        clsParams = getTypeParameters(cls)
     else:
         clsParams = None
 
     line(out_h)
-    line(out_h, indent, 'class t_%s {', names[-1])
+    line(out_h, indent, 'class %st_%s {', _dll_export, names[-1])
     line(out_h, indent, 'public:')
     line(out_h, indent + 1, 'PyObject_HEAD')
     line(out_h, indent + 1, '%s object;', cppname(names[-1]))
@@ -497,16 +576,22 @@ def python(env, out_h, out, cls, superCls, names, superNames,
 
     line(out_h, indent + 1, 'static PyObject *wrap_Object(const %s&);',
          cppname(names[-1]))
-    if clsParams:
-        line(out_h, indent + 1, 'static PyObject *wrap_Object(const %s&, %s);',
-             cppname(names[-1]),
-             ', '.join(['PyTypeObject *'] * len(clsParams)))
     line(out_h, indent + 1, 'static PyObject *wrap_jobject(const jobject&);')
+    if clsParams:
+        _clsParams = ', '.join(['PyTypeObject *'] * len(clsParams))
+        line(out_h, indent + 1, 'static PyObject *wrap_Object(const %s&, %s);',
+             cppname(names[-1]), _clsParams)
+        line(out_h, indent + 1, 'static PyObject *wrap_jobject(const jobject&, %s);', _clsParams)
     line(out_h, indent + 1, 'static void install(PyObject *module);')
     line(out_h, indent + 1, 'static void initialize(PyObject *module);')
     line(out_h, indent, '};')
 
-    iterator = findClass('java/util/Iterator')
+    if env.java_version >= '1.5':
+        iterable = findClass('java/lang/Iterable')
+        iterator = findClass('java/util/Iterator')
+    else:
+        iterable = iterator = None
+
     enumeration = findClass('java/util/Enumeration')
 
     while indent:
@@ -539,6 +624,10 @@ def python(env, out_h, out, cls, superCls, names, superNames,
 
     line(out, indent, 'static PyObject *t_%s_cast_(PyTypeObject *type, PyObject *arg);', names[-1])
     line(out, indent, 'static PyObject *t_%s_instance_(PyTypeObject *type, PyObject *arg);', names[-1])
+    if clsParams:
+        line(out, indent,
+             'static PyObject *t_%s_of_(t_%s *self, PyObject *args);',
+             names[-1], names[-1])
 
     if constructors:
         line(out, indent, 'static int t_%s_init_(t_%s *self, PyObject *args, PyObject *kwds);', names[-1], names[-1])
@@ -640,7 +729,9 @@ def python(env, out_h, out, cls, superCls, names, superNames,
         method = methods[0]
         modifiers = method.getModifiers()
         if name == 'iterator' and iteratorMethod is None:
-            if (not method.getParameterTypes() and
+            if (iterable is not None and
+                not method.getParameterTypes() and
+                iterable.isAssignableFrom(cls) and
                 iterator.isAssignableFrom(method.getReturnType())):
                 iteratorMethod = method
         elif name == 'next' and nextMethod is None:
@@ -676,7 +767,9 @@ def python(env, out_h, out, cls, superCls, names, superNames,
         method = methods[0]
         modifiers = method.getModifiers()
         if name == 'iterator' and iteratorMethod is None:
-            if (not method.getParameterTypes() and
+            if (iterable is not None and
+                not method.getParameterTypes() and
+                iterable.isAssignableFrom(cls) and
                 iterator.isAssignableFrom(method.getReturnType())):
                 iteratorMethod = method
                 iteratorExt = True
@@ -775,6 +868,9 @@ def python(env, out_h, out, cls, superCls, names, superNames,
          'DECLARE_METHOD(t_%s, cast_, METH_O | METH_CLASS),', names[-1])
     line(out, indent + 1,
          'DECLARE_METHOD(t_%s, instance_, METH_O | METH_CLASS),', names[-1])
+    if clsParams:
+        line(out, indent + 1,
+             'DECLARE_METHOD(t_%s, of_, METH_VARARGS),', names[-1])
 
     for name, methods in allMethods:
         modifiers = methods[0].getModifiers()
@@ -809,7 +905,7 @@ def python(env, out_h, out, cls, superCls, names, superNames,
         else:
             tp_iter = '((PyObject *(*)(t_%s *)) get_%siterator<t_%s>)' %(names[-1], clsParams and 'generic_' or '', names[-1])
         tp_iternext = '0'
-    elif nextMethod and iterator.isAssignableFrom(cls):
+    elif nextMethod and iterable is not None and iterator.isAssignableFrom(cls):
         tp_iter = 'PyObject_SelfIter'
         returnName = typename(nextMethod.getReturnType(), cls, False)
         ns, sep, n = rpartition(returnName, '::')
@@ -910,7 +1006,7 @@ def python(env, out_h, out, cls, superCls, names, superNames,
         line(out, indent, "{")
         line(out, indent + 1, "PyObject *obj = t_%s::wrap_Object(object);",
              names[-1])
-        line(out, indent + 1, "if (obj != Py_None)")
+        line(out, indent + 1, "if (obj != NULL && obj != Py_None)")
         line(out, indent + 1, "{")
         line(out, indent + 2, "t_%s *self = (t_%s *) obj;",
              names[-1], names[-1])
@@ -918,6 +1014,27 @@ def python(env, out_h, out, cls, superCls, names, superNames,
         for clsParam in clsParams:
             line(out, indent + 2, "self->parameters[%d] = %s;",
                  i, clsParam.getName())
+            i += 1
+        line(out, indent + 1, "}")
+        line(out, indent + 1, "return obj;");
+        line(out, indent, "}")
+
+        line(out)
+        line(out, indent, 
+             "PyObject *t_%s::wrap_jobject(const jobject& object, %s)",
+             cppname(names[-1]), ', '.join(clsArgs))
+        line(out, indent, "{")
+        line(out, indent + 1, "PyObject *obj = t_%s::wrap_jobject(object);",
+             names[-1])
+        line(out, indent + 1, "if (obj != NULL && obj != Py_None)")
+        line(out, indent + 1, "{")
+        line(out, indent + 2, "t_%s *self = (t_%s *) obj;",
+             names[-1], names[-1])
+        i = 0;
+        for clsParam in clsParams:
+            line(out, indent + 2, "self->parameters[%d] = %s;",
+                 i, clsParam.getName())
+            i += 1
         line(out, indent + 1, "}")
         line(out, indent + 1, "return obj;");
         line(out, indent, "}")
@@ -925,23 +1042,31 @@ def python(env, out_h, out, cls, superCls, names, superNames,
     line(out)
     line(out, indent, 'void t_%s::install(PyObject *module)', names[-1])
     line(out, indent, '{')
-    line(out, indent + 1, 'installType(&%s$$Type, module, "%s", %d);',
+    line(out, indent + 1, 'installType(&PY_TYPE(%s), module, "%s", %d);',
          names[-1], rename or names[-1], isExtension and 1 or 0)
     for inner in cls.getDeclaredClasses():
         if inner in typeset:
             if Modifier.isStatic(inner.getModifiers()):
                 innerName = inner.getName().split('.')[-1]
-                line(out, indent + 1, 'PyDict_SetItemString(%s$$Type.tp_dict, "%s", make_descriptor(&%s$$Type));',
+                line(out, indent + 1, 'PyDict_SetItemString(PY_TYPE(%s).tp_dict, "%s", make_descriptor(&PY_TYPE(%s)));',
                      names[-1], innerName[len(names[-1])+1:], innerName)
     line(out, indent, '}')
 
     line(out)
     line(out, indent, 'void t_%s::initialize(PyObject *module)', names[-1])
     line(out, indent, '{')
-    line(out, indent + 1, 'PyDict_SetItemString(%s$$Type.tp_dict, "class_", make_descriptor(%s::initializeClass, %s));',
+    line(out, indent + 1, 'PyDict_SetItemString(PY_TYPE(%s).tp_dict, "class_", make_descriptor(%s::initializeClass, %s));',
          names[-1], cppname(names[-1]), generics and 1 or 0)
-    line(out, indent + 1, 'PyDict_SetItemString(%s$$Type.tp_dict, "wrapfn_", make_descriptor(t_%s::wrap_jobject));',
-         names[-1], names[-1])
+
+    if is_unboxed(cls.getName()):
+        wrapfn_ = "unbox%s" %(names[-1])
+        boxfn_ = "box%s" %(names[-1])
+    else:
+        wrapfn_ = "t_%s::wrap_jobject" %(names[-1])
+        boxfn_ = "boxObject"
+
+    line(out, indent + 1, 'PyDict_SetItemString(PY_TYPE(%s).tp_dict, "wrapfn_", make_descriptor(%s));', names[-1], wrapfn_)
+    line(out, indent + 1, 'PyDict_SetItemString(PY_TYPE(%s).tp_dict, "boxfn_", make_descriptor(%s));', names[-1], boxfn_)
 
     if isExtension:
         line(out, indent + 1, 'jclass cls = %s::initializeClass();',
@@ -966,7 +1091,7 @@ def python(env, out_h, out, cls, superCls, names, superNames,
         fieldName = field.getName()
         value = '%s::%s' %(cppname(names[-1]), cppname(fieldName))
         value = fieldValue(cls, value, fieldType)
-        line(out, indent + 1, 'PyDict_SetItemString(%s$$Type.tp_dict, "%s", make_descriptor(%s));',
+        line(out, indent + 1, 'PyDict_SetItemString(PY_TYPE(%s).tp_dict, "%s", make_descriptor(%s));',
              names[-1], fieldName, value)
     line(out, indent, '}')
 
@@ -985,6 +1110,20 @@ def python(env, out_h, out, cls, superCls, names, superNames,
     line(out, indent + 1, 'Py_RETURN_TRUE;')
     line(out, indent, '}')
 
+    if clsParams:
+        line(out)
+        line(out, indent,
+             'static PyObject *t_%s_of_(t_%s *self, PyObject *args)',
+             names[-1], names[-1])
+        line(out, indent, '{')
+        line(out, indent + 1,
+             'if (!parseArg(args, "T", %d, &(self->parameters)))',
+             len(clsParams))
+        line(out, indent + 2, 'Py_RETURN_SELF;');
+        line(out, indent + 1,
+             'return PyErr_SetArgsError((PyObject *) self, "of_", args);')
+        line(out, indent, '}')
+
     if constructors:
         line(out)
         line(out, indent, 'static int t_%s_init_(t_%s *self, PyObject *args, PyObject *kwds)', names[-1], names[-1])
@@ -1070,7 +1209,7 @@ def python(env, out_h, out, cls, superCls, names, superNames,
                     line(out, indent + 1, 'return callSuper(type, "%s"%s, %d);',
                          name, args, cardinality)
                 else:
-                    line(out, indent + 1, 'return callSuper(&%s$$Type, (PyObject *) self, "%s"%s, %d);',
+                    line(out, indent + 1, 'return callSuper(&PY_TYPE(%s), (PyObject *) self, "%s"%s, %d);',
                          names[-1], name, args, cardinality)
             else:
                 line(out, indent + 1, 'PyErr_SetArgsError(%s, "%s"%s);',
@@ -1250,7 +1389,11 @@ def python(env, out_h, out, cls, superCls, names, superNames,
              not returnType.isPrimitive() and '((jobject) NULL)' or '')
         line(out, indent + 1, 'OBJ_CALL(result = self->object.%s((%s) n));',
              methodName, typename(method.getParameterTypes()[0], cls, False))
-        line(out, indent + 1, returnValue(cls, returnType, 'result'))
+        if generics:
+            line(out, indent + 1, returnValue(cls, returnType, 'result',
+                                              method.getGenericReturnType()))
+        else:
+            line(out, indent + 1, returnValue(cls, returnType, 'result'))
         line(out, indent, '}')
 
     while indent:
@@ -1333,9 +1476,13 @@ def package(out, allInOne, cppdir, namespace, names):
     if not names:
         line(out, indent + 1, 'return env;')
         indent -= 1
-        line(out, indent + 1, '} catch (JCCEnv::exception e) {')
-        line(out, indent + 2, 'PyErr_SetJavaError(e.throwable);')
-        line(out, indent + 2, 'return NULL;')
+        line(out, indent + 1, '} catch (int e) {')
+        line(out, indent + 2, 'switch(e) {')
+        line(out, indent + 2, '  case _EXC_JAVA:')
+        line(out, indent + 3, 'return PyErr_SetJavaError();')
+        line(out, indent + 2, '  default:')
+        line(out, indent + 3, 'throw;')
+        line(out, indent + 2, '}')
         line(out, indent + 1, '}')
 
     line(out, indent, '}')
@@ -1353,7 +1500,8 @@ def package(out, allInOne, cppdir, namespace, names):
         package(out, allInOne, cppdir, entries, names + (name,))
 
 
-def module(out, allInOne, classes, cppdir, moduleName, shared, generics):
+def module(out, allInOne, classes, imports, cppdir, moduleName,
+           shared, generics):
 
     extname = '_%s' %(moduleName)
     line(out, 0, '#include <Python.h>')
@@ -1364,11 +1512,15 @@ def module(out, allInOne, classes, cppdir, moduleName, shared, generics):
         out_init = file(os.path.join(cppdir, '__init__.cpp'), 'w')
     namespaces = {}
     for cls in classes:
-        namespace = namespaces
-        classNames = cls.getName().split('.')
-        for className in classNames[:-1]:
-            namespace = namespace.setdefault(className, {})
-        namespace[classNames[-1]] = True
+        for importset in imports.itervalues():
+            if cls in importset:
+                break
+        else:
+            namespace = namespaces
+            classNames = cls.getName().split('.')
+            for className in classNames[:-1]:
+                namespace = namespace.setdefault(className, {})
+            namespace[classNames[-1]] = True
     if allInOne:
         package(out_init, True, cppdir, namespaces, ())
         out_init.close()
@@ -1378,7 +1530,7 @@ def module(out, allInOne, classes, cppdir, moduleName, shared, generics):
     line(out)
     line(out, 0, 'PyObject *initJCC(PyObject *module);')
     line(out, 0, 'void __install__(PyObject *module);')
-    line(out, 0, 'extern PyTypeObject JObject$$Type, ConstVariableDescriptor$$Type, FinalizerClass$$Type, FinalizerProxy$$Type;')
+    line(out, 0, 'extern PyTypeObject PY_TYPE(JObject), PY_TYPE(ConstVariableDescriptor), PY_TYPE(FinalizerClass), PY_TYPE(FinalizerProxy);')
     line(out, 0, 'extern void _install_jarray(PyObject *);')
     line(out)
     line(out, 0, 'extern "C" {')
@@ -1402,8 +1554,9 @@ def module(out, allInOne, classes, cppdir, moduleName, shared, generics):
 
 
 def compile(env, jccPath, output, moduleName, install, dist, debug, jars,
-            version, prefix, root, install_dir, use_distutils,
-            shared, compiler, modules, wininst, arch, generics):
+            version, prefix, root, install_dir, home_dir, use_distutils,
+            shared, compiler, modules, wininst, find_jvm_dll, arch, generics,
+            resources, imports):
 
     try:
         if use_distutils:
@@ -1432,6 +1585,9 @@ def compile(env, jccPath, output, moduleName, install, dist, debug, jars,
         line(out, 0, "import os, sys")
         line(out)
         line(out, 0, "if sys.platform == 'win32':")
+        if find_jvm_dll:
+            line(out, 1, "from jcc.windows import add_jvm_dll_directory_to_path")
+            line(out, 1, "add_jvm_dll_directory_to_path()")
         line(out, 1, "import jcc, %s", extname)
         line(out, 0, "else:")
         line(out, 1, "import %s", extname)
@@ -1444,6 +1600,25 @@ def compile(env, jccPath, output, moduleName, install, dist, debug, jars,
     for jar in jars:
         shutil.copy2(jar, modulePath)
         package_data.append(os.path.basename(jar))
+    if resources:
+        def copytree(src, dst):
+            _dst = os.path.join(modulePath, dst)
+            if not os.path.exists(_dst):
+                os.mkdir(_dst)
+            for name in os.listdir(src):
+                if name.startswith('.'):
+                    continue
+                _src = os.path.join(src, name)
+                if os.path.islink(_src):
+                    continue
+                _dst = os.path.join(dst, name)
+                if os.path.isdir(_src):
+                    copytree(_src, _dst)
+                else:
+                    shutil.copy2(_src, os.path.join(modulePath, _dst))
+                    package_data.append(_dst)
+        for resource in resources:
+            copytree(resource, os.path.split(resource)[-1])
 
     packages = [moduleName]
     if modules:
@@ -1479,8 +1654,8 @@ def compile(env, jccPath, output, moduleName, install, dist, debug, jars,
     line(out, 1, 'def getJavaException(self):')
     line(out, 2, 'return self.args[0]')
     line(out, 1, 'def __str__(self):')
-    line(out, 2, 'writer = %s.StringWriter()', extname)
-    line(out, 2, 'self.getJavaException().printStackTrace(%s.PrintWriter(writer))', extname)
+    line(out, 2, 'writer = StringWriter()')
+    line(out, 2, 'self.getJavaException().printStackTrace(PrintWriter(writer))')
     line(out, 2, 'return "\\n".join((super(JavaError, self).__str__(), "    Java stacktrace:", str(writer)))')
     line(out)
     line(out, 0, 'class InvalidArgsError(Exception):')
@@ -1500,11 +1675,15 @@ def compile(env, jccPath, output, moduleName, install, dist, debug, jars,
          extname, extname, extname)
 
     line(out)
+    for import_ in imports:
+        line(out, 0, 'from %s._%s import *', import_.__name__, import_.__name__)
     line(out, 0, 'from %s import *', extname)
     out.close()
 
     includes = [os.path.join(output, extname),
                 os.path.join(jccPath, 'sources')]
+    for import_ in imports:
+        includes.append(os.path.join(import_.__dir__, 'include'))
 
     sources = ['JObject.cpp', 'JArray.cpp', 'functions.cpp', 'types.cpp']
     if not shared:
@@ -1514,6 +1693,25 @@ def compile(env, jccPath, output, moduleName, install, dist, debug, jars,
 	shutil.copy2(os.path.join(jccPath, 'sources', source),
                      os.path.join(output, extname))
 
+    if shared:
+        def copytree(src, dst):
+            _dst = os.path.join(modulePath, dst)
+            if not os.path.exists(_dst):
+                os.mkdir(_dst)
+            for name in os.listdir(src):
+                if name.startswith('.'):
+                    continue
+                _src = os.path.join(src, name)
+                if os.path.islink(_src):
+                    continue
+                _dst = os.path.join(dst, name)
+                if os.path.isdir(_src):
+                    copytree(_src, _dst)
+                elif name.endswith('.h'):
+                    shutil.copy2(_src, os.path.join(modulePath, _dst))
+                    package_data.append(_dst)
+        copytree(os.path.join(output, extname), 'include')
+
     sources = []
     for path, dirs, names in os.walk(os.path.join(output, extname)):
         for name in names:
@@ -1525,20 +1723,22 @@ def compile(env, jccPath, output, moduleName, install, dist, debug, jars,
     includes[0:0] = INCLUDES
     compile_args = CFLAGS
     link_args = LFLAGS
-    defines=['PYTHON']
+
+    defines=[('PYTHON', None),
+             ('JCC_VER', '"%s"' %(JCC_VER))]
+    if shared:
+        defines.append(('_jcc_shared', None))
+    if generics:
+        defines.append(('_java_generics', None))
 
     if compiler:
         script_args.append('--compiler=%s' %(compiler))
 
-    if shared:
-        defines.append('_jcc_shared')
-    script_args.append('--define=%s' %(','.join(defines)))
-
     if debug:
         script_args.append('--debug')
         compile_args += DEBUG_CFLAGS
     elif sys.platform == 'win32':
-        pass
+	pass
     elif sys.platform == 'sunos5':
         link_args.append('-Wl,-s')
     else:
@@ -1552,6 +1752,8 @@ def compile(env, jccPath, output, moduleName, install, dist, debug, jars,
         script_args.append('--root=%s' % root)
     if install_dir:
         script_args.append('--install-lib=%s' % install_dir)
+    if home_dir:
+        script_args.append('--home=%s' % home_dir)
 
     if dist:
         if wininst:
@@ -1565,12 +1767,10 @@ def compile(env, jccPath, output, moduleName, install, dist, debug, jars,
         'extra_compile_args': compile_args,
         'extra_link_args': link_args,
         'include_dirs': includes,
-        'sources': sources
+        'sources': sources,
+        'define_macros': defines
     }
 
-    if generics:
-        args['define_macros'] = [('_java_generics', None)]
-
     if shared:
         shlibdir = os.path.dirname(os.path.dirname(_jcc.__file__))
         if sys.platform == 'darwin':   # distutils no good with -R
@@ -1585,9 +1785,30 @@ def compile(env, jccPath, output, moduleName, install, dist, debug, jars,
             args['extra_link_args'] += ['-Wl,-rpath', shlibdir]
             args['library_dirs'] = [shlibdir]
             args['libraries'] = ['jcc']
+            args['extra_link_args'] += [
+                getattr(import_, "_%s" %(import_.__name__)).__file__
+                for import_ in imports
+            ]
         elif sys.platform == 'win32':
-            jcclib = 'jcc%s.lib' %(debug and '_d' or '')
-            args['extra_link_args'] += [os.path.join(shlibdir, 'jcc', jcclib)]
+            _d = debug and '_d' or ''
+            libdir = os.path.join(modulePath, 'lib')
+            if not os.path.exists(libdir):
+                os.mkdir(libdir)
+            extlib = os.path.join('lib', "%s%s.lib" %(extname, _d))
+            package_data.append(extlib)
+            args['extra_link_args'] += [
+                os.path.join(shlibdir, 'jcc', 'jcc%s.lib' %(_d)),
+                ' '.join(IMPLIB_LFLAGS) %(os.path.join(modulePath, extlib))
+            ]
+            args['libraries'] = [
+                os.path.join(import_.__dir__, 'lib',
+                             '_%s%s' %(import_.__name__, _d))
+                for import_ in imports
+            ]
+            args['define_macros'] += [
+                ("_dll_%s" %(import_.__name__), '__declspec(dllimport)')
+                for import_ in imports
+            ] + [("_dll_%s" %(moduleName), '__declspec(dllexport)')]
         else:
             raise NotImplementedError, "shared mode on %s" %(sys.platform)
 
